package kayttoliittyma;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

import org.junit.Before;
import org.junit.Test;

public class ProxyTest {
    
    private static interface Target {
        public void mVoid();
        public int mInt();
        public int mIntInt(int x);
        public int mIntString(String x);
        public String mStringIntString(int x, String s);
        
        public int overloaded(int x);
        public String overloaded(String x);
    }
    
    private Target target;
    
    private Proxy<Target> proxy;
    
    @Before
    public void setUp() {
        target = mock(Target.class);
        proxy = new Proxy<Target>(target);
    }
    
    @Test
    public void shouldForwardParameterlessVoidMethod() {
        proxy.callMethod(Void.TYPE, "mVoid");
        verify(target).mVoid();
    }
    
    @Test
    public void shouldForwardParameterlessIntMethod() {
        when(target.mInt()).thenReturn(3);
        int result = proxy.callMethod(Integer.TYPE, "mInt");
        assertEquals(3, result);
    }
    
    @Test
    public void shouldForwardMethodWithOnePrimitiveParameter() {
        when(target.mIntInt(7)).thenReturn(3);
        int result = proxy.callMethod(Integer.TYPE, "mIntInt", 7);
        assertEquals(3, result);
    }
    
    @Test
    public void shouldForwardMethodWithOneObjectParameter() {
        when(target.mIntString("xoo")).thenReturn(3);
        int result = proxy.callMethod(Integer.TYPE, "mIntString", "xoo");
        assertEquals(3, result);
    }
    
    @Test
    public void shouldForwardMethodWithTwoParameters() {
        when(target.mStringIntString(7, "xoo")).thenReturn("it works");
        String result = proxy.callMethod(String.class, "mStringIntString", 7, "xoo");
        assertEquals("it works", result);
    }
    
    @Test
    public void shouldForwardOverloadedMethodBasedOnParameters() {
        when(target.overloaded(7)).thenReturn(3);
        when(target.overloaded("x")).thenReturn("y");
        int result1 = proxy.callMethod(Integer.TYPE, "overloaded", 7);
        String result2 = proxy.callMethod(String.class, "overloaded", "x");
        assertEquals(3, result1);
        assertEquals("y", result2);
    }
    
    @Test(expected = ProxyException.class)
    public void shouldThrowProxyExceptionIfMethodNotFound() {
        proxy.callMethod(String.class, "nonexistent");
    }
    
    @Test(expected = ProxyException.class)
    public void shouldThrowProxyExceptionIfTooManyParameters() {
        proxy.callMethod(String.class, "mInt", 3);
    }
    
    @Test(expected = ProxyException.class)
    public void shouldThrowProxyExceptionIfTooFewParameters() {
        proxy.callMethod(String.class, "mIntInt");
    }
    
    @Test(expected = ProxyException.class)
    public void shouldNotFindMethodIfReturnTypeMismatches() {
        proxy.callMethod(String.class, "mInt");
    }
    
    @Test
    public void shouldDescribeMissingMethodInException() {
        ProxyException ex = null;
        try {
            proxy.callMethod(Integer.TYPE, "nonexistent", "foo", 3, 4.5);
        } catch (ProxyException e) {
            ex = e;
        }
        assertNotNull(ex);
        assertTrue(
            "Invalid exception message: " + ex.getMessage(),
            ex.getMessage().contains("public int nonexistent(String, int, double)"));
    }
}
